#include "TimSynth.h"

#pragma udata access seqtiming_data
near unsigned char microstep;

near unsigned char step_pos;
near unsigned char num_steps;
near unsigned char current_step;

near unsigned char bar_pos;
near unsigned char num_bars;
near unsigned char current_bar;

near unsigned char new_step;
near union _u16_u8x2 seq_t;
near unsigned short t0_reload;
near unsigned long short seqtiming_speed;

#pragma udata
unsigned char ts;
unsigned short seqtiming_speed_8step;

static unsigned long short speed;
static unsigned long bps;

void seqtiming_init(unsigned short bpm)
{
	T0CON = 0b00000011;

	num_steps = 8;
	num_bars = 8;

	seqtiming_restart(bpm);

	INTCONbits.TMR0IF = 0;
}

// bpm in 100ths of BPM, 200 = 2.00 bpm
void seqtiming_restart(unsigned short bpm)
{
	if(bpm)
	{
		if(bpm == 1)
		{
			goto SEQUENCER_RESTART;
		}

RECALC_TIMER:
		seqtiming_set_bpm(bpm);

SEQUENCER_RESTART:
		T0CONbits.TMR0ON = 0;

		microstep = 4;
		new_step = 0;
		step_pos = 0x01;
		bar_pos = 0x01;
		med_led = 0x01;
		dim_led = 0x01;

		current_bar = 0;
		current_step = 0;

		T0CONbits.TMR0ON = 1;
	}
}

void seqtiming_pause(void)
{
	T0CONbits.TMR0ON = 0;
}

void seqtiming_continue(void)
{
	T0CONbits.TMR0ON = 1;
}

void seqtiming_calculate_tspeed(unsigned short bpm)
{
	if(bpm)
	{
		if(bpm >= BPM_MIN)
		{
			if(bpm <= BPM_MAX)
			{
				speed = bpm;
			}
			else
			{
				speed = BPM_MAX;
			}
		}
		else
		{
			speed = BPM_MIN;
		}

	
		if(speed <= 375)			// 3.75 bpm
		{
			ts = 0b00000111;
			bps = 11718750 / speed;
		}
		else if(speed <= 750)			// 7.5 bpm
		{
			ts = 0b00000110;
			bps = 23437500 / speed;
		}
		else if(speed <= 1500)		// 15 bpm
		{
			ts = 0b00000101;
			bps = 46875000 / speed;
		}
		else if(speed <= 3000)		// 30 bpm
		{
			ts = 0b00000100;
			bps = 93750000 / speed;
		}
		else if(speed <= 6000)		// 60 bpm
		{
			ts = 0b00000011;
			bps = 187500000 / speed;
		}
		else if(speed <= 12000)	// 120 bpm
		{
			ts = 0b00000010;
			bps = 375000000 / speed;
		}
		else						// 240 bpm
		{
			ts = 0b00000001;
			bps = 750000000 / speed;
		}
		
		speed = 0x10000 - bps;
	}
}

void seqtiming_load_tspeed(unsigned short tspeed)
{
	if(tspeed)
	{
		speed = tspeed;
		t0_reload = speed;
	
		INTCONbits.TMR0IE = 0;
	
		T0CON = ts;
	
		seq_t.lowbyte = TMR0L;
		seq_t.hibyte = TMR0H;
		seq_t.val += t0_reload;
		
		TMR0H = seq_t.hibyte;
		TMR0L = seq_t.lowbyte;

		T0CONbits.TMR0ON = 1;
	
		INTCONbits.TMR0IE = 1;
	}
}

void seqtiming_set_bpm(unsigned short bpm)
{
	static float scalc;
	
	if(bpm)
	{
		if(bpm >= BPM_MIN)
		{
			if(bpm <= BPM_MAX)
			{
				seqtiming_speed = bpm;
			}
			else
			{
				seqtiming_speed = BPM_MAX;
			}
		}
		else
		{
			seqtiming_speed = BPM_MIN;
		}

		seqtiming_speed_8step = seqtiming_speed;

		scalc = seqtiming_speed;
		
		scalc *= ((float)num_steps / 8.0);
		
		seqtiming_speed = scalc;	
	
		seqtiming_calculate_tspeed(seqtiming_speed);
		seqtiming_load_tspeed(speed);
	}
}

void seqtiming_faster(unsigned short bpm_increase)
{
	if(bpm_increase)
	{
		seqtiming_set_bpm(seqtiming_speed_8step+bpm_increase);
	}
}

void seqtiming_slower(unsigned short bpm_decrease)
{
	if(bpm_decrease)
	{
		seqtiming_set_bpm(seqtiming_speed_8step-bpm_decrease);
	}
}

void seqtiming_stop(void)
{
	seqtiming_pause();
	
	INTCONbits.TMR0IE = 0;
	INTCONbits.TMR0IF = 0;
	
	seq_t.val = t0_reload;
	TMR0H = seq_t.hibyte;
	TMR0L = seq_t.lowbyte;
}

void seqtiming_start(void)
{
	T0CONbits.TMR0ON = 1;
	INTCONbits.TMR0IE = 1;
}

void seqtiming_reset_steps(void)
{
	seqtiming_stop();
	current_step = 0;
	step_pos = 0;
	seqtiming_start();
}

void seqtiming_reset_bars(void)
{
	seqtiming_stop();
	current_bar = 0;
	bar_pos = 0;
	seqtiming_start();
}

void seqtiming_set_num_steps(unsigned char nsteps)
{
	if((nsteps >= 1) && (nsteps <= STEP_MAX))
	{
		if(nsteps != num_steps)
		{
			seqtiming_stop();
			num_steps = nsteps;
			seqtiming_reset_steps();
			seqtiming_set_bpm(seqtiming_speed_8step);
			seqtiming_start();
		}
	}
}

void seqtiming_set_num_bars(unsigned char nbars)
{
	if((nbars >= 1) && (nbars <= BAR_MAX))
	{
		if(nbars != num_bars)
		{
			seqtiming_stop();
			seqtiming_reset_bars();
			num_bars = nbars;
			seqtiming_start();
		}
	}
}
